@@ -35,6 +35,7 @@
 #include "common.h"
 #include "simutil.h"
 #include "util.h"
+#include "storage.h"
 
 #define NETWORK_REGISTRATION_INTERFACE "org.ofono.NetworkRegistration"
 #define NETWORK_OPERATOR_INTERFACE "org.ofono.NetworkOperator"
@@ -44,11 +45,19 @@
 #define NETWORK_REGISTRATION_FLAG_HOME_SHOW_PLMN 0x4
 #define NETWORK_REGISTRATION_FLAG_ROAMING_SHOW_SPN 0x8
 
-#define AUTO_REGISTER 1
+enum network_registration_mode {
+	NETWORK_REGISTRATION_MODE_AUTO = 0,
+	NETWORK_REGISTRATION_MODE_MANUAL = 1,
+	NETWORK_REGISTRATION_MODE_OFF = 2,
+	NETWORK_REGISTRATION_MODE_MANUAL_AUTO = 4
+};
 
 /* How often we update the operator list, in seconds */
 #define OPERATOR_LIST_UPDATE_TIME 300
 
+#define SETTINGS_STORE "netreg"
+#define SETTINGS_GROUP "Settings"
+
 static GSList *g_drivers = NULL;
 
 /* 27.007 Section 7.3 <stat> */
@@ -64,6 +73,7 @@ struct ofono_netreg {
 	int location;
 	int cellid;
 	int technology;
+	int mode;
 	char *base_station;
 	struct network_operator_data *current_operator;
 	GSList *operator_list;
@@ -76,6 +86,8 @@ struct ofono_netreg {
 	struct sim_eons *eons;
 	gint opscan_source;
 	struct ofono_sim *sim;
+	GKeyFile *settings;
+	char *imsi;
 	struct ofono_watchlist *status_watches;
 	const struct ofono_netreg_driver *driver;
 	void *driver_data;
@@ -98,11 +110,29 @@ static void registration_status_callback(const struct ofono_error *error,
 					void *data);
 
 struct network_operator_data {
-	struct ofono_network_operator *info;
+	char name[OFONO_MAX_OPERATOR_NAME_LENGTH + 1];
+	char mcc[OFONO_MAX_MCC_LENGTH + 1];
+	char mnc[OFONO_MAX_MNC_LENGTH + 1];
+	int status;
+	unsigned int techs;
 	const struct sim_eons_operator_info *eons_info;
 	struct ofono_netreg *netreg;
 };
 
+static const char *registration_mode_to_string(int mode)
+{
+	switch (mode) {
+	case NETWORK_REGISTRATION_MODE_AUTO:
+		return "auto";
+	case NETWORK_REGISTRATION_MODE_MANUAL:
+		return "manual";
+	case NETWORK_REGISTRATION_MODE_OFF:
+		return "off";
+	default:
+		return "unknown";
+	}
+}
+
 static inline const char *network_operator_status_to_string(int status)
 {
 	switch (status) {
@@ -117,6 +147,58 @@ static inline const char *network_operator_status_to_string(int status)
 	return "unknown";
 }
 
+static char **network_operator_technologies(struct network_operator_data *opd)
+{
+	unsigned int ntechs = 0;
+	char **techs;
+	unsigned int i;
+
+	for (i = 0; i < sizeof(opd->techs); i++) {
+		if (opd->techs & (1 << i))
+			ntechs += 1;
+	}
+
+	techs = g_new0(char *, ntechs + 1);
+	ntechs = 0;
+
+	for (i = 0; i < sizeof(opd->techs); i++) {
+		if (!(opd->techs & (1 << i)))
+			continue;
+
+		techs[ntechs++] = g_strdup(registration_tech_to_string(i));
+	}
+
+	return techs;
+}
+
+static void set_registration_mode(struct ofono_netreg *netreg, int mode)
+{
+	DBusConnection *conn;
+	const char *strmode;
+	const char *path;
+
+	if (netreg->mode == mode)
+		return;
+
+	netreg->mode = mode;
+
+	if (netreg->settings) {
+		g_key_file_set_integer(netreg->settings, SETTINGS_GROUP,
+					"Mode", netreg->mode);
+		storage_sync(netreg->imsi, SETTINGS_STORE, netreg->settings);
+	}
+
+	strmode = registration_mode_to_string(mode);
+
+	conn = ofono_dbus_get_connection();
+	path = __ofono_atom_get_path(netreg->atom);
+
+	ofono_dbus_signal_property_changed(conn, path,
+						NETWORK_REGISTRATION_INTERFACE,
+						"Mode", DBUS_TYPE_STRING,
+						&strmode);
+}
+
 static void register_callback(const struct ofono_error *error, void *data)
 {
 	struct ofono_netreg *netreg = data;
@@ -190,12 +272,11 @@ static void network_operator_populate_registered(struct ofono_netreg *netreg,
 	i = 0;
 	for (l = netreg->operator_list; l; l = l->next) {
 		struct network_operator_data *opd = l->data;
-		struct ofono_network_operator *op = opd->info;
 		int j;
 
 		for (j = 0; children[j]; j++) {
 			sscanf(children[j], "%3[0-9]%[0-9]", mcc, mnc);
-			if (!strcmp(op->mcc, mcc) && !strcmp(op->mnc, mnc)) {
+			if (!strcmp(opd->mcc, mcc) && !strcmp(opd->mnc, mnc)) {
 				(*network_operators)[i] =
 					g_try_new(char, op_path_len);
 				snprintf((*network_operators)[i], op_path_len,
@@ -208,23 +289,54 @@ static void network_operator_populate_registered(struct ofono_netreg *netreg,
 	dbus_free_string_array(children);
 }
 
+static struct network_operator_data *
+	network_operator_create(const struct ofono_network_operator *op)
+{
+	struct network_operator_data *opd;
+
+	opd = g_new0(struct network_operator_data, 1);
+
+	memcpy(&opd->name, op->name, sizeof(opd->name));
+	memcpy(&opd->mcc, op->mcc, sizeof(opd->mcc));
+	memcpy(&opd->mnc, op->mnc, sizeof(opd->mnc));
+
+	opd->status = op->status;
+
+	if (op->tech != -1)
+		opd->techs |= 1 << op->tech;
+
+	return opd;
+}
+
 static void network_operator_destroy(gpointer userdata)
 {
 	struct network_operator_data *op = userdata;
 
-	g_free(op->info);
 	g_free(op);
 }
 
 static gint network_operator_compare(gconstpointer a, gconstpointer b)
 {
 	const struct network_operator_data *opda = a;
-	const struct ofono_network_operator *opa = opda->info;
 	const struct ofono_network_operator *opb = b;
 
 	int comp1;
 	int comp2;
 
+	comp1 = strcmp(opda->mcc, opb->mcc);
+	comp2 = strcmp(opda->mnc, opb->mnc);
+
+	return comp1 != 0 ? comp1 : comp2;
+}
+
+static gint network_operator_data_compare(gconstpointer a, gconstpointer b)
+{
+	const struct network_operator_data *opa = a;
+	const struct network_operator_data *opb = b;
+
+	int comp1;
+	int comp2;
+
 	comp1 = strcmp(opa->mcc, opb->mcc);
 	comp2 = strcmp(opa->mnc, opb->mnc);
 
@@ -232,13 +344,14 @@ static gint network_operator_compare(gconstpointer a, gconstpointer b)
 }
 
 static inline const char *network_operator_build_path(struct ofono_netreg *netreg,
-				const struct ofono_network_operator *oper)
+							const char *mcc,
+							const char *mnc)
 {
 	static char path[256];
 
 	snprintf(path, sizeof(path), "%s/operator/%s%s",
 			__ofono_atom_get_path(netreg->atom),
-			oper->mcc, oper->mnc);
+			mcc, mnc);
 
 	return path;
 }
@@ -260,53 +373,53 @@ static void network_operator_emit_available_operators(struct ofono_netreg *netre
 	g_strfreev(network_operators);
 }
 
-static void set_network_operator_status(struct ofono_netreg *netreg,
-					struct network_operator_data *opd,
+static void set_network_operator_status(struct network_operator_data *opd,
 					int status)
 {
-	struct ofono_network_operator *op = opd->info;
 	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_netreg *netreg = opd->netreg;
 	const char *status_str;
 	const char *path;
 
-	if (op->status == status)
+	if (opd->status == status)
 		return;
 
-	op->status = status;
+	opd->status = status;
 
 	status_str = network_operator_status_to_string(status);
-	path = network_operator_build_path(netreg, op);
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 
 	ofono_dbus_signal_property_changed(conn, path, NETWORK_OPERATOR_INTERFACE,
 						"Status", DBUS_TYPE_STRING,
 						&status_str);
 }
 
-static void set_network_operator_technology(struct ofono_netreg *netreg,
-					struct network_operator_data *opd,
-					int tech)
+static void set_network_operator_techs(struct network_operator_data *opd,
+					unsigned int techs)
 {
-	struct ofono_network_operator *op = opd->info;
 	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *tech_str;
+	struct ofono_netreg *netreg = opd->netreg;
+	char **technologies;
 	const char *path;
 
-	if (op->tech == tech)
+	if (opd->techs == techs)
 		return;
 
-	op->tech = tech;
-	tech_str = registration_tech_to_string(tech);
-	path = network_operator_build_path(netreg, op);
+	opd->techs = techs;
+	technologies = network_operator_technologies(opd);
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 
-	ofono_dbus_signal_property_changed(conn, path, NETWORK_OPERATOR_INTERFACE,
-						"Technology", DBUS_TYPE_STRING,
-						&tech_str);
+	ofono_dbus_signal_array_property_changed(conn, path,
+						NETWORK_REGISTRATION_INTERFACE,
+						"Technologies",
+						DBUS_TYPE_STRING,
+						&technologies);
+	g_strfreev(technologies);
 }
 
 static char *get_operator_display_name(struct ofono_netreg *netreg)
 {
-	struct network_operator_data *current = netreg->current_operator;
-	struct ofono_network_operator *op;
+	struct network_operator_data *opd = netreg->current_operator;
 	const char *plmn;
 	static char name[1024];
 	int len = sizeof(name);
@@ -316,16 +429,14 @@ static char *get_operator_display_name(struct ofono_netreg *netreg)
 	 * PLMN or roaming and on configuration bits from the SIM, all
 	 * together there are four cases to consider.  */
 
-	if (!current) {
+	if (!opd) {
 		g_strlcpy(name, "", len);
 		return name;
 	}
 
-	op = current->info;
-
-	plmn = op->name;
-	if (current->eons_info && current->eons_info->longname)
-		plmn = current->eons_info->longname;
+	plmn = opd->name;
+	if (opd->eons_info && opd->eons_info->longname)
+		plmn = opd->eons_info->longname;
 
 	if (!netreg->spname || strlen(netreg->spname) == 0) {
 		g_strlcpy(name, plmn, len);
@@ -335,7 +446,8 @@ static char *get_operator_display_name(struct ofono_netreg *netreg)
 	if (netreg->status == NETWORK_REGISTRATION_STATUS_REGISTERED)
 		home_or_spdi = TRUE;
 	else
-		home_or_spdi = sim_spdi_lookup(netreg->spdi, op->mcc, op->mnc);
+		home_or_spdi = sim_spdi_lookup(netreg->spdi,
+							opd->mcc, opd->mnc);
 
 	if (home_or_spdi)
 		if (netreg->flags & NETWORK_REGISTRATION_FLAG_HOME_SHOW_PLMN)
@@ -355,20 +467,19 @@ static char *get_operator_display_name(struct ofono_netreg *netreg)
 	return name;
 }
 
-static void set_network_operator_name(struct ofono_netreg *netreg,
-					struct network_operator_data *opd,
+static void set_network_operator_name(struct network_operator_data *opd,
 					const char *name)
 {
-	struct ofono_network_operator *op = opd->info;
 	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_netreg *netreg = opd->netreg;
 	const char *path;
 	const char *operator;
 
-	if (!strncmp(op->name, name, OFONO_MAX_OPERATOR_NAME_LENGTH))
+	if (!strncmp(opd->name, name, OFONO_MAX_OPERATOR_NAME_LENGTH))
 		return;
 
-	strncpy(op->name, name, OFONO_MAX_OPERATOR_NAME_LENGTH);
-	op->name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
+	strncpy(opd->name, name, OFONO_MAX_OPERATOR_NAME_LENGTH);
+	opd->name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
 
 	/* If we have Enhanced Operator Name info on the SIM, we always use
 	 * that, so do not need to emit the signal here
@@ -376,7 +487,7 @@ static void set_network_operator_name(struct ofono_netreg *netreg,
 	if (opd->eons_info && opd->eons_info->longname)
 		return;
 
-	path = network_operator_build_path(netreg, op);
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 
 	ofono_dbus_signal_property_changed(conn, path, NETWORK_OPERATOR_INTERFACE,
 					"Name", DBUS_TYPE_STRING, &name);
@@ -393,11 +504,11 @@ static void set_network_operator_name(struct ofono_netreg *netreg,
 	}
 }
 
-static void set_network_operator_eons_info(struct ofono_netreg *netreg,
-				struct network_operator_data *opd,
+static void set_network_operator_eons_info(struct network_operator_data *opd,
 				const struct sim_eons_operator_info *eons_info)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_netreg *netreg = opd->netreg;
 	const struct sim_eons_operator_info *old_eons_info = opd->eons_info;
 	const char *path;
 	const char *oldname;
@@ -408,18 +519,18 @@ static void set_network_operator_eons_info(struct ofono_netreg *netreg,
 	if (!old_eons_info && !eons_info)
 		return;
 
-	path = network_operator_build_path(netreg, opd->info);
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 	opd->eons_info = eons_info;
 
 	if (old_eons_info && old_eons_info->longname)
 		oldname = old_eons_info->longname;
 	else
-		oldname = opd->info->name;
+		oldname = opd->name;
 
 	if (eons_info && eons_info->longname)
 		newname = eons_info->longname;
 	else
-		newname = opd->info->name;
+		newname = opd->name;
 
 	if (oldname != newname && strcmp(oldname, newname)) {
 		ofono_dbus_signal_property_changed(conn, path,
@@ -464,9 +575,9 @@ static DBusMessage *network_operator_get_properties(DBusConnection *conn,
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 
-	const char *name = opd->info->name;
+	const char *name = opd->name;
 	const char *status =
-		network_operator_status_to_string(opd->info->status);
+		network_operator_status_to_string(opd->status);
 
 	if (opd->eons_info && opd->eons_info->longname)
 		name = opd->eons_info->longname;
@@ -485,24 +596,26 @@ static DBusMessage *network_operator_get_properties(DBusConnection *conn,
 
 	ofono_dbus_dict_append(&dict, "Status", DBUS_TYPE_STRING, &status);
 
-	if (*opd->info->mcc != '\0') {
-		const char *mcc = opd->info->mcc;
+	if (*opd->mcc != '\0') {
+		const char *mcc = opd->mcc;
 		ofono_dbus_dict_append(&dict, "MobileCountryCode",
 					DBUS_TYPE_STRING, &mcc);
 	}
 
-	if (*opd->info->mnc != '\0') {
-		const char *mnc = opd->info->mnc;
+	if (*opd->mnc != '\0') {
+		const char *mnc = opd->mnc;
 		ofono_dbus_dict_append(&dict, "MobileNetworkCode",
 					DBUS_TYPE_STRING, &mnc);
 	}
 
-	if (opd->info->tech != -1) {
-		const char *technology =
-			registration_tech_to_string(opd->info->tech);
+	if (opd->techs != 0) {
+		char **technologies = network_operator_technologies(opd);
 
-		ofono_dbus_dict_append(&dict, "Technology", DBUS_TYPE_STRING,
-					&technology);
+		ofono_dbus_dict_append_array(&dict, "Technologies",
+						DBUS_TYPE_STRING,
+						&technologies);
+
+		g_strfreev(technologies);
 	}
 
 	if (opd->eons_info && opd->eons_info->info) {
@@ -520,8 +633,8 @@ static DBusMessage *network_operator_get_properties(DBusConnection *conn,
 static DBusMessage *network_operator_register(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
-	struct network_operator_data *op = data;
-	struct ofono_netreg *netreg = op->netreg;
+	struct network_operator_data *opd = data;
+	struct ofono_netreg *netreg = opd->netreg;
 
 	if (netreg->flags & NETWORK_REGISTRATION_FLAG_PENDING)
 		return __ofono_error_busy(msg);
@@ -532,9 +645,11 @@ static DBusMessage *network_operator_register(DBusConnection *conn,
 	netreg->flags |= NETWORK_REGISTRATION_FLAG_PENDING;
 	netreg->pending = dbus_message_ref(msg);
 
-	netreg->driver->register_manual(netreg, op->info,
+	netreg->driver->register_manual(netreg, opd->mcc, opd->mnc,
 					register_callback, netreg);
 
+	set_registration_mode(netreg, NETWORK_REGISTRATION_MODE_MANUAL);
+
 	return NULL;
 }
 
@@ -550,58 +665,40 @@ static GDBusSignalTable network_operator_signals[] = {
 	{ }
 };
 
-static struct network_operator_data *
-	network_operator_dbus_register(struct ofono_netreg *netreg,
-					const struct ofono_network_operator *op,
-					enum operator_status status)
+static gboolean network_operator_dbus_register(struct ofono_netreg *netreg,
+					struct network_operator_data *opd)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
 	const char *path;
-	struct network_operator_data *opd = NULL;
 
-	path = network_operator_build_path(netreg, op);
-
-	opd = g_try_new(struct network_operator_data, 1);
-
-	if (!opd)
-		goto err;
-
-	opd->info = g_memdup(op, sizeof(struct ofono_network_operator));
-
-	if (opd->info == NULL)
-		goto err;
-
-	opd->info->status = status;
-	opd->netreg = netreg;
-	opd->eons_info = NULL;
-
-	if (netreg->eons)
-		opd->eons_info = sim_eons_lookup(netreg->eons,
-							op->mcc, op->mnc);
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 
 	if (!g_dbus_register_interface(conn, path, NETWORK_OPERATOR_INTERFACE,
 					network_operator_methods,
 					network_operator_signals,
 					NULL, opd,
-					network_operator_destroy))
-		goto err;
+					network_operator_destroy)) {
+		ofono_error("Could not register NetworkOperator %s", path);
+		return FALSE;
+	}
 
-	return opd;
+	opd->netreg = netreg;
+	opd->eons_info = NULL;
 
-err:
-	if (opd)
-		network_operator_destroy(opd);
+	if (netreg->eons)
+		opd->eons_info = sim_eons_lookup(netreg->eons,
+							opd->mcc, opd->mnc);
 
-	ofono_error("Could not register NetworkOperator %s", path);
-
-	return NULL;
+	return TRUE;
 }
 
 static gboolean network_operator_dbus_unregister(struct ofono_netreg *netreg,
 						struct network_operator_data *opd)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
-	const char *path = network_operator_build_path(netreg, opd->info);
+	const char *path;
+
+	path = network_operator_build_path(netreg, opd->mcc, opd->mnc);
 
 	return g_dbus_unregister_interface(conn, path,
 					NETWORK_OPERATOR_INTERFACE);
@@ -617,6 +714,7 @@ static DBusMessage *network_get_properties(DBusConnection *conn,
 
 	const char *status = registration_status_to_string(netreg->status);
 	const char *operator;
+	const char *mode = registration_mode_to_string(netreg->mode);
 
 	char **network_operators;
 
@@ -631,6 +729,7 @@ static DBusMessage *network_get_properties(DBusConnection *conn,
 					&dict);
 
 	ofono_dbus_dict_append(&dict, "Status", DBUS_TYPE_STRING, &status);
+	ofono_dbus_dict_append(&dict, "Mode", DBUS_TYPE_STRING, &mode);
 
 	if (netreg->location != -1) {
 		dbus_uint16_t location = netreg->location;
@@ -694,6 +793,8 @@ static DBusMessage *network_register(DBusConnection *conn,
 
 	netreg->driver->register_auto(netreg, register_callback, netreg);
 
+	set_registration_mode(netreg, NETWORK_REGISTRATION_MODE_AUTO);
+
 	return NULL;
 }
 
@@ -713,6 +814,8 @@ static DBusMessage *network_deregister(DBusConnection *conn,
 
 	netreg->driver->deregister(netreg, register_callback, netreg);
 
+	set_registration_mode(netreg, NETWORK_REGISTRATION_MODE_OFF);
+
 	return NULL;
 }
 
@@ -839,6 +942,10 @@ void __ofono_netreg_set_base_station_name(struct ofono_netreg *netreg,
 	const char *path = __ofono_atom_get_path(netreg->atom);
 	const char *base_station = name ? name : "";
 
+	/* Cell ID changed, but we don't have a cell name, nothing to do */
+	if (netreg->base_station == NULL && name == NULL)
+		return;
+
 	if (netreg->base_station)
 		g_free(netreg->base_station);
 
@@ -894,17 +1001,20 @@ static void notify_status_watches(struct ofono_netreg *netreg)
 	struct ofono_watchlist_item *item;
 	GSList *l;
 	ofono_netreg_status_notify_cb_t notify;
-	struct ofono_network_operator *op = NULL;
+	const char *mcc = NULL;
+	const char *mnc = NULL;
 
-	if (netreg->current_operator)
-		op = netreg->current_operator->info;
+	if (netreg->current_operator) {
+		mcc = netreg->current_operator->mcc;
+		mnc = netreg->current_operator->mnc;
+	}
 
 	for (l = netreg->status_watches->items; l; l = l->next) {
 		item = l->data;
 		notify = item->notify;
 
 		notify(netreg->status, netreg->location, netreg->cellid,
-			netreg->technology, op, item->notify_data);
+			netreg->technology, mcc, mnc, item->notify_data);
 	}
 }
 
@@ -946,6 +1056,36 @@ void ofono_netreg_status_notify(struct ofono_netreg *netreg, int status,
 	notify_status_watches(netreg);
 }
 
+static GSList *compress_operator_list(const struct ofono_network_operator *list,
+					int total)
+{
+	GSList *oplist = 0;
+	GSList *o;
+	int i;
+	struct network_operator_data *opd;
+
+	for (i = 0; i < total; i++) {
+		o = NULL;
+
+		if (oplist)
+			o = g_slist_find_custom(oplist, &list[i],
+						network_operator_compare);
+
+		if (!o) {
+			opd = network_operator_create(&list[i]);
+			oplist = g_slist_prepend(oplist, opd);
+		} else if (o && list[i].tech != -1) {
+			opd = o->data;
+			opd->techs |= 1 << list[i].tech;
+		}
+	}
+
+	if (oplist)
+		oplist = g_slist_reverse(oplist);
+
+	return oplist;
+}
+
 static void operator_list_callback(const struct ofono_error *error, int total,
 					const struct ofono_network_operator *list,
 					void *data)
@@ -953,7 +1093,8 @@ static void operator_list_callback(const struct ofono_error *error, int total,
 	struct ofono_netreg *netreg = data;
 	GSList *n = NULL;
 	GSList *o;
-	int i;
+	GSList *compressed;
+	GSList *c;
 	gboolean need_to_emit = FALSE;
 
 	netreg->flags &= ~NETWORK_REGISTRATION_FLAG_REQUESTING_OPLIST;
@@ -963,19 +1104,18 @@ static void operator_list_callback(const struct ofono_error *error, int total,
 		return;
 	}
 
-	for (i = 0; i < total; i++) {
-		o = g_slist_find_custom(netreg->operator_list, &list[i],
-					network_operator_compare);
+	compressed = compress_operator_list(list, total);
+
+	for (c = compressed; c; c = c->next) {
+		struct network_operator_data *copd = c->data;
+
+		o = g_slist_find_custom(netreg->operator_list, copd,
+					network_operator_data_compare);
 
 		if (o) { /* Update and move to a new list */
-			set_network_operator_status(netreg, o->data,
-							list[i].status);
-
-			set_network_operator_technology(netreg, o->data,
-							list[i].tech);
-
-			set_network_operator_name(netreg, o->data,
-							list[i].name);
+			set_network_operator_status(o->data, copd->status);
+			set_network_operator_techs(o->data, copd->techs);
+			set_network_operator_name(o->data, copd->name);
 
 			n = g_slist_prepend(n, o->data);
 			netreg->operator_list =
@@ -984,17 +1124,22 @@ static void operator_list_callback(const struct ofono_error *error, int total,
 			/* New operator */
 			struct network_operator_data *opd;
 
-			opd = network_operator_dbus_register(netreg, &list[i],
-								list[i].status);
+			opd = g_memdup(copd,
+					sizeof(struct network_operator_data));
 
-			if (!opd)
+			if (!network_operator_dbus_register(netreg, opd)) {
+				g_free(opd);
 				continue;
+			}
 
 			n = g_slist_prepend(n, opd);
 			need_to_emit = TRUE;
 		}
 	}
 
+	g_slist_foreach(compressed, (GFunc)g_free, NULL);
+	g_slist_free(compressed);
+
 	if (n)
 		n = g_slist_reverse(n);
 
@@ -1037,7 +1182,7 @@ static void current_operator_callback(const struct ofono_error *error,
 	if (netreg->current_operator &&
 		(!current ||
 			network_operator_compare(netreg->current_operator, current)))
-		set_network_operator_status(netreg, netreg->current_operator,
+		set_network_operator_status(netreg->current_operator,
 						OPERATOR_STATUS_AVAILABLE);
 
 	if (current)
@@ -1045,11 +1190,16 @@ static void current_operator_callback(const struct ofono_error *error,
 					network_operator_compare);
 
 	if (op) {
-		set_network_operator_status(netreg, op->data,
-						OPERATOR_STATUS_CURRENT);
-		set_network_operator_technology(netreg, op->data,
-						current->tech);
-		set_network_operator_name(netreg, op->data, current->name);
+		struct network_operator_data *opd = op->data;
+		unsigned int techs = opd->techs;
+
+		if (current->tech != -1) {
+			techs |= 1 << current->tech;
+			set_network_operator_techs(opd, techs);
+		}
+
+		set_network_operator_status(opd, OPERATOR_STATUS_CURRENT);
+		set_network_operator_name(opd, current->name);
 
 		if (netreg->current_operator == op->data)
 			return;
@@ -1061,11 +1211,12 @@ static void current_operator_callback(const struct ofono_error *error,
 	if (current) {
 		struct network_operator_data *opd;
 
-		opd = network_operator_dbus_register(netreg, current,
-						OPERATOR_STATUS_CURRENT);
+		opd = network_operator_create(current);
 
-		if (!opd)
+		if (!network_operator_dbus_register(netreg, opd)) {
+			g_free(opd);
 			return;
+		}
 
 		netreg->current_operator = opd;
 		netreg->operator_list = g_slist_append(netreg->operator_list,
@@ -1126,9 +1277,10 @@ static void init_registration_status(const struct ofono_error *error,
 					signal_strength_callback, netreg);
 	}
 
-	if (AUTO_REGISTER &&
+	if (netreg->mode == NETWORK_REGISTRATION_MODE_AUTO &&
 		(status == NETWORK_REGISTRATION_STATUS_NOT_REGISTERED ||
-		 status == NETWORK_REGISTRATION_STATUS_DENIED))
+			status == NETWORK_REGISTRATION_STATUS_DENIED ||
+			status == NETWORK_REGISTRATION_STATUS_UNKNOWN))
 		netreg->driver->register_auto(netreg, register_callback, netreg);
 }
 
@@ -1204,10 +1356,9 @@ optimize:
 		struct network_operator_data *opd = l->data;
 		const struct sim_eons_operator_info *eons_info;
 
-		eons_info = sim_eons_lookup(netreg->eons, opd->info->mcc,
-						opd->info->mnc);
+		eons_info = sim_eons_lookup(netreg->eons, opd->mcc, opd->mnc);
 
-		set_network_operator_eons_info(netreg, opd, eons_info);
+		set_network_operator_eons_info(opd, eons_info);
 	}
 }
 
@@ -1266,7 +1417,7 @@ static void sim_spdi_read_cb(int ok, int length, int record,
 		const char *operator;
 
 		if (!sim_spdi_lookup(netreg->spdi,
-					current->info->mcc, current->info->mnc))
+					current->mcc, current->mnc))
 			return;
 
 		operator = get_operator_display_name(netreg);
@@ -1375,8 +1526,7 @@ int ofono_netreg_get_technology(struct ofono_netreg *netreg)
 	return netreg->technology;
 }
 
-const struct ofono_network_operator *
-	ofono_netreg_get_operator(struct ofono_netreg *netreg)
+const char *ofono_netreg_get_mcc(struct ofono_netreg *netreg)
 {
 	if (netreg == NULL)
 		return NULL;
@@ -1384,7 +1534,18 @@ const struct ofono_network_operator *
 	if (netreg->current_operator == NULL)
 		return NULL;
 
-	return netreg->current_operator->info;
+	return netreg->current_operator->mcc;
+}
+
+const char *ofono_netreg_get_mnc(struct ofono_netreg *netreg)
+{
+	if (netreg == NULL)
+		return NULL;
+
+	if (netreg->current_operator == NULL)
+		return NULL;
+
+	return netreg->current_operator->mnc;
 }
 
 int ofono_netreg_driver_register(const struct ofono_netreg_driver *d)
@@ -1433,6 +1594,15 @@ static void netreg_unregister(struct ofono_atom *atom)
 		netreg->base_station = NULL;
 	}
 
+	if (netreg->settings) {
+		storage_close(netreg->imsi, SETTINGS_STORE,
+				netreg->settings, TRUE);
+
+		g_free(netreg->imsi);
+		netreg->imsi = NULL;
+		netreg->settings = NULL;
+	}
+
 	g_dbus_unregister_interface(conn, path,
 					NETWORK_REGISTRATION_INTERFACE);
 	ofono_modem_remove_interface(modem, NETWORK_REGISTRATION_INTERFACE);
@@ -1503,6 +1673,33 @@ struct ofono_netreg *ofono_netreg_create(struct ofono_modem *modem,
 	return netreg;
 }
 
+static void netreg_load_settings(struct ofono_netreg *netreg)
+{
+	const char *imsi;
+	int mode;
+
+	imsi = ofono_sim_get_imsi(netreg->sim);
+
+	if (!imsi)
+		return;
+
+	netreg->settings = storage_open(imsi, SETTINGS_STORE);
+
+	if (netreg->settings == NULL)
+		return;
+
+	netreg->imsi = g_strdup(imsi);
+
+	mode = g_key_file_get_integer(netreg->settings, SETTINGS_GROUP,
+					"Mode", NULL);
+
+	if (mode >= 0 && mode <= 2)
+		netreg->mode = mode;
+
+	g_key_file_set_integer(netreg->settings, SETTINGS_GROUP,
+				"Mode", netreg->mode);
+}
+
 void ofono_netreg_register(struct ofono_netreg *netreg)
 {
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -1539,6 +1736,8 @@ void ofono_netreg_register(struct ofono_netreg *netreg)
 		/* Assume that if sim atom exists, it is ready */
 		netreg->sim = __ofono_atom_get_data(sim_atom);
 
+		netreg_load_settings(netreg);
+
 		ofono_sim_read(netreg->sim, SIM_EFPNN_FILEID,
 				OFONO_SIM_FILE_STRUCTURE_FIXED,
 				sim_pnn_read_cb, netreg);
